"""CLI package facade."""

from .main import (
    Config,
    SvgOptions,
    build_client_edges,
    build_device_index,
    build_node_type_map,
    build_topology,
    debug_dump_devices,
    fetch_clients,
    fetch_devices,
    group_devices_by_type,
    main,
    normalize_devices,
    render_legend,
    render_mermaid,
    render_svg,
    resolve_themes,
    write_output,
)

__all__ = [
    "Config",
    "SvgOptions",
    "build_client_edges",
    "build_device_index",
    "build_node_type_map",
    "build_topology",
    "debug_dump_devices",
    "fetch_clients",
    "fetch_devices",
    "group_devices_by_type",
    "main",
    "normalize_devices",
    "render_legend",
    "render_mermaid",
    "render_svg",
    "resolve_themes",
    "write_output",
]



"""Output helpers for files and stdout."""

from __future__ import annotations

import sys
from pathlib import Path


def write_output(content: str, *, output_path: str | None, stdout: bool) -> None:
    if output_path:
        Path(output_path).write_text(content, encoding="utf-8")
    if stdout or not output_path:
        sys.stdout.write(content)



"""CLI entry point."""

from __future__ import annotations

import argparse
import logging

from ..adapters.config import Config
from ..adapters.unifi import fetch_clients, fetch_devices
from ..io.debug import debug_dump_devices
from ..io.export import write_output
from ..model.topology import (
    Device,
    build_client_edges,
    build_device_index,
    build_node_type_map,
    build_topology,
    group_devices_by_type,
    normalize_devices,
)
from ..render.mermaid import render_legend, render_mermaid
from ..render.mermaid_theme import MermaidTheme
from ..render.svg import SvgOptions, render_svg
from ..render.svg_theme import SvgTheme
from ..render.theme import resolve_themes

logger = logging.getLogger(__name__)


def _load_dotenv(env_file: str | None = None) -> None:
    try:
        from dotenv import load_dotenv
    except ImportError:
        logger.info("python-dotenv not installed; skipping .env loading")
        return
    load_dotenv(dotenv_path=env_file) if env_file else load_dotenv()


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Generate network maps from UniFi LLDP data, as mermaid or SVG"
    )
    _add_source_args(parser.add_argument_group("Source"))
    _add_functional_args(parser.add_argument_group("Functional"))
    _add_mermaid_args(parser.add_argument_group("Mermaid"))
    _add_svg_args(parser.add_argument_group("SVG"))
    _add_general_render_args(parser.add_argument_group("Output"))
    _add_debug_args(parser.add_argument_group("Debug"))
    return parser


def _add_source_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--site", default=None, help="UniFi site name (overrides UNIFI_SITE)")
    parser.add_argument(
        "--env-file",
        default=None,
        help="Path to .env file (overrides default .env discovery)",
    )


def _add_functional_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--include-ports", action="store_true", help="Include port labels in edges")
    parser.add_argument(
        "--include-clients",
        action="store_true",
        help="Include active clients as leaf nodes",
    )
    parser.add_argument(
        "--only-unifi", action="store_true", help="Only include neighbors that are UniFi devices"
    )


def _add_mermaid_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--direction", default="TB", choices=["LR", "TB"], help="Mermaid direction")
    parser.add_argument(
        "--group-by-type",
        action="store_true",
        help="Group nodes by gateway/switch/ap in Mermaid subgraphs",
    )
    parser.add_argument(
        "--legend-only",
        action="store_true",
        help="Render only the legend as a separate Mermaid graph",
    )


def _add_svg_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument("--svg-width", type=int, default=None, help="SVG width override")
    parser.add_argument("--svg-height", type=int, default=None, help="SVG height override")
    parser.add_argument("--theme-file", default=None, help="Path to theme YAML file")


def _add_general_render_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        "--format",
        default="mermaid",
        choices=["mermaid", "svg", "svg-iso"],
        help="Output format",
    )
    parser.add_argument(
        "--markdown",
        action="store_true",
        help="Wrap output in a Markdown mermaid code fence for notes tools like Obsidian",
    )
    parser.add_argument("--output", default=None, help="Output file path")
    parser.add_argument("--stdout", action="store_true", help="Write output to stdout")


def _add_debug_args(parser: argparse._ArgumentGroup) -> None:
    parser.add_argument(
        "--debug-dump",
        action="store_true",
        help="Dump gateway and sample device data to stderr for debugging",
    )
    parser.add_argument(
        "--debug-sample",
        type=int,
        default=2,
        help="Number of non-gateway devices to include in debug dump (default: 2)",
    )


def _parse_args(argv: list[str] | None) -> argparse.Namespace:
    parser = _build_parser()
    return parser.parse_args(argv)


def _load_config(args: argparse.Namespace) -> Config | None:
    try:
        _load_dotenv(args.env_file)
        return Config.from_env(env_file=args.env_file)
    except ValueError as exc:
        logging.error(str(exc))
        return None


def _resolve_site(args: argparse.Namespace, config: Config) -> str:
    return args.site or config.site


def _render_legend_only(args: argparse.Namespace, mermaid_theme: MermaidTheme) -> str:
    content = render_legend(theme=mermaid_theme)
    if args.markdown:
        content = f"""```mermaid
{content}```
"""
    return content


def _build_topology_data(
    args: argparse.Namespace, config: Config, site: str
) -> tuple[list[Device], list[str], object]:
    raw_devices = list(fetch_devices(config, site=site, detailed=True))
    devices = normalize_devices(raw_devices)
    if args.debug_dump:
        debug_dump_devices(raw_devices, devices, sample_count=max(0, args.debug_sample))
    groups_for_rank = group_devices_by_type(devices)
    gateways = groups_for_rank.get("gateway", [])
    topology = build_topology(
        devices,
        include_ports=args.include_ports,
        only_unifi=args.only_unifi,
        gateways=gateways,
    )
    return devices, gateways, topology


def _build_edges_with_clients(
    args: argparse.Namespace,
    edges: list,
    devices: list[Device],
    config: Config,
    site: str,
) -> tuple[list, list | None]:
    clients = None
    if args.include_clients:
        clients = list(fetch_clients(config, site=site))
        device_index = build_device_index(devices)
        edges = edges + build_client_edges(clients, device_index, include_ports=args.include_ports)
    return edges, clients


def _select_edges(topology: object) -> tuple[list, bool]:
    if topology.tree_edges:
        return topology.tree_edges, True
    logging.warning("No gateway found for hierarchy; rendering raw edges.")
    return topology.raw_edges, False


def _render_mermaid_output(
    args: argparse.Namespace,
    devices: list[Device],
    topology: object,
    config: Config,
    site: str,
    mermaid_theme: MermaidTheme,
) -> str:
    edges, _has_tree = _select_edges(topology)
    edges, clients = _build_edges_with_clients(args, edges, devices, config, site)
    groups = None
    group_order = None
    if args.group_by_type:
        groups = group_devices_by_type(devices)
        group_order = ["gateway", "switch", "ap", "other"]
    content = render_mermaid(
        edges,
        direction=args.direction,
        groups=groups,
        group_order=group_order,
        node_types=build_node_type_map(devices, clients),
        theme=mermaid_theme,
    )
    if args.markdown:
        content = f"""```mermaid
{content}```
"""
    return content


def _render_svg_output(
    args: argparse.Namespace,
    devices: list[Device],
    topology: object,
    config: Config,
    site: str,
    svg_theme: SvgTheme,
) -> str:
    edges, _has_tree = _select_edges(topology)
    edges, clients = _build_edges_with_clients(args, edges, devices, config, site)
    options = SvgOptions(width=args.svg_width, height=args.svg_height)
    if args.format == "svg-iso":
        from ..render.svg import render_svg_isometric

        return render_svg_isometric(
            edges,
            node_types=build_node_type_map(devices, clients),
            options=options,
            theme=svg_theme,
        )
    return render_svg(
        edges,
        node_types=build_node_type_map(devices, clients),
        options=options,
        theme=svg_theme,
    )


def main(argv: list[str] | None = None) -> int:
    logging.basicConfig(level=logging.INFO, format="%(levelname)s %(name)s: %(message)s")
    args = _parse_args(argv)
    config = _load_config(args)
    if config is None:
        return 2
    site = _resolve_site(args, config)
    mermaid_theme, svg_theme = resolve_themes(args.theme_file)

    if args.legend_only:
        content = _render_legend_only(args, mermaid_theme)
        write_output(content, output_path=args.output, stdout=args.stdout)
        return 0

    try:
        devices, _gateways, topology = _build_topology_data(args, config, site)
    except Exception as exc:
        logging.error("Failed to build topology: %s", exc)
        return 1

    if args.format == "mermaid":
        content = _render_mermaid_output(args, devices, topology, config, site, mermaid_theme)
    elif args.format in {"svg", "svg-iso"}:
        content = _render_svg_output(args, devices, topology, config, site, svg_theme)
    else:
        logging.error("Unsupported format: %s", args.format)
        return 2

    write_output(content, output_path=args.output, stdout=args.stdout)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Mermaid theming helpers."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class MermaidTheme:
    node_gateway: tuple[str, str]
    node_switch: tuple[str, str]
    node_ap: tuple[str, str]
    node_client: tuple[str, str]
    node_other: tuple[str, str]
    poe_link: str
    poe_link_width: int
    poe_link_arrow: str
    standard_link: str
    standard_link_width: int
    standard_link_arrow: str


DEFAULT_THEME = MermaidTheme(
    node_gateway=("#ffe3b3", "#d98300"),
    node_switch=("#d6ecff", "#3a7bd5"),
    node_ap=("#d7f5e7", "#27ae60"),
    node_client=("#f2e5ff", "#7f3fbf"),
    node_other=("#eeeeee", "#8f8f8f"),
    poe_link="#1e88e5",
    poe_link_width=2,
    poe_link_arrow="none",
    standard_link="#2ecc71",
    standard_link_width=2,
    standard_link_arrow="none",
)


def class_defs(theme: MermaidTheme = DEFAULT_THEME) -> list[str]:
    return [
        f"  classDef node_gateway fill:{theme.node_gateway[0]},stroke:{theme.node_gateway[1]},stroke-width:1px;",
        f"  classDef node_switch fill:{theme.node_switch[0]},stroke:{theme.node_switch[1]},stroke-width:1px;",
        f"  classDef node_ap fill:{theme.node_ap[0]},stroke:{theme.node_ap[1]},stroke-width:1px;",
        f"  classDef node_client fill:{theme.node_client[0]},stroke:{theme.node_client[1]},stroke-width:1px;",
        f"  classDef node_other fill:{theme.node_other[0]},stroke:{theme.node_other[1]},stroke-width:1px;",
        "  classDef node_legend font-size:10px;",
    ]



"""Mermaid diagram rendering."""

from __future__ import annotations

from collections.abc import Iterable

from ..model.topology import Edge
from .mermaid_theme import DEFAULT_THEME, MermaidTheme, class_defs


def _escape(label: str) -> str:
    return label.replace('"', '\\"')


def _slugify(value: str) -> str:
    normalized = []
    for ch in value.strip():
        if ch.isalnum():
            normalized.append(ch.lower())
        else:
            normalized.append("_")
    slug = "".join(normalized).strip("_")
    if not slug or slug[0].isdigit():
        slug = f"n_{slug}" if slug else "n"
    return slug


def _build_id_map(edges: Iterable[Edge], nodes: Iterable[str]) -> dict[str, str]:
    id_map: dict[str, str] = {}
    used: set[str] = set()

    def assign(name: str) -> None:
        if name in id_map:
            return
        base = _slugify(name)
        candidate = base
        counter = 2
        while candidate in used:
            candidate = f"{base}_{counter}"
            counter += 1
        id_map[name] = candidate
        used.add(candidate)

    for node in nodes:
        assign(node)
    for edge in edges:
        assign(edge.left)
        assign(edge.right)

    return id_map


def _node_ref(name: str, node_id: str) -> str:
    return f'{node_id}["{_escape(name)}"]'


def render_mermaid(
    edges: Iterable[Edge],
    direction: str = "LR",
    *,
    groups: dict[str, list[str]] | None = None,
    group_order: list[str] | None = None,
    node_types: dict[str, str] | None = None,
    theme: MermaidTheme = DEFAULT_THEME,
) -> str:
    edge_list = list(edges)
    group_nodes: list[str] = []
    if groups:
        for members in groups.values():
            group_nodes.extend(members)
    id_map = _build_id_map(edge_list, group_nodes)
    lines = [f"graph {direction}"]
    poe_links: list[int] = []
    link_index = 0
    if groups:
        ordered = group_order or list(groups.keys())
        for group_name in ordered:
            members = groups.get(group_name, [])
            if not members:
                continue
            group_id = _slugify(f"group_{group_name}")
            label = group_name.replace("_", " ").title()
            lines.append(f'  subgraph {group_id}["{_escape(label)}"];')
            for member in members:
                lines.append(f"    {_node_ref(member, id_map[member])};")
            lines.append("  end")
    use_node_labels = not groups
    for edge in edge_list:
        if use_node_labels:
            left = _node_ref(edge.left, id_map[edge.left])
            right = _node_ref(edge.right, id_map[edge.right])
        else:
            left = id_map[edge.left]
            right = id_map[edge.right]
        if edge.label:
            label = _escape(edge.label)
            lines.append(f'  {left} ---|"{label}"| {right};')
        else:
            lines.append(f"  {left} --- {right};")
        if edge.poe:
            poe_links.append(link_index)
        link_index += 1
    if node_types:
        class_map = {
            "gateway": "node_gateway",
            "switch": "node_switch",
            "ap": "node_ap",
            "client": "node_client",
            "other": "node_other",
        }
        if node_types:
            for name, node_type in node_types.items():
                class_name = class_map.get(node_type, "node_other")
                node_id = id_map.get(name)
                if node_id:
                    lines.append(f"  class {node_id} {class_name};")
        lines.extend(class_defs(theme))
    for index in poe_links:
        lines.append(
            "  linkStyle "
            f"{index} stroke:{theme.poe_link},stroke-width:{theme.poe_link_width}px,"
            f"arrowhead:{theme.poe_link_arrow};"
        )
    return "\n".join(lines) + "\n"


def render_legend(theme: MermaidTheme = DEFAULT_THEME) -> str:
    lines = [
        "graph TB",
        '  subgraph legend["Legend"];',
        '    legend_gateway["Gateway"];',
        '    legend_switch["Switch"];',
        '    legend_ap["AP"];',
        '    legend_client["Client"];',
        '    legend_other["Other"];',
        '    legend_poe_a["PoE Link A"];',
        '    legend_poe_b["PoE Link B"];',
        '    legend_no_poe_a["Link A"];',
        '    legend_no_poe_b["Link B"];',
        "    legend_poe_a ---|⚡| legend_poe_b;",
        "    legend_no_poe_a --- legend_no_poe_b;",
        "    linkStyle 0 arrowhead:none;",
        "    linkStyle 1 arrowhead:none;",
        "  end",
        "  class legend_gateway node_gateway;",
        "  class legend_switch node_switch;",
        "  class legend_ap node_ap;",
        "  class legend_client node_client;",
        "  class legend_other node_other;",
        "  class legend_poe_a node_legend;",
        "  class legend_poe_b node_legend;",
        "  class legend_no_poe_a node_legend;",
        "  class legend_no_poe_b node_legend;",
    ]
    lines.extend(class_defs(theme))
    lines.append(
        "  linkStyle 0 "
        f"stroke:{theme.poe_link},stroke-width:{theme.poe_link_width}px,"
        f"arrowhead:{theme.poe_link_arrow};"
    )
    lines.append(
        "  linkStyle 1 "
        f"stroke:{theme.standard_link},stroke-width:{theme.standard_link_width}px,"
        f"arrowhead:{theme.standard_link_arrow};"
    )
    return "\n".join(lines) + "\n"



"""UniFi API integration."""

from __future__ import annotations

import hashlib
import logging
import os
import pickle
import time
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

from .config import Config

if TYPE_CHECKING:
    from unifi_controller_api import UnifiController

logger = logging.getLogger(__name__)


def _cache_dir() -> Path:
    return Path(os.environ.get("UNIFI_CACHE_DIR", ".cache/unifi_network_maps"))


def _cache_ttl_seconds() -> int:
    value = os.environ.get("UNIFI_CACHE_TTL_SECONDS", "").strip()
    if not value:
        return 3600
    if value.isdigit():
        return int(value)
    logger.warning("Invalid UNIFI_CACHE_TTL_SECONDS value: %s", value)
    return 3600


def _cache_key(*parts: str) -> str:
    digest = hashlib.sha256("|".join(parts).encode("utf-8")).hexdigest()
    return digest[:24]


def _load_cache(path: Path, ttl_seconds: int) -> object | None:
    if ttl_seconds <= 0 or not path.exists():
        return None
    try:
        payload = pickle.loads(path.read_bytes())
    except Exception as exc:
        logger.debug("Failed to read cache %s: %s", path, exc)
        return None
    timestamp = payload.get("timestamp")
    if not isinstance(timestamp, int | float):
        return None
    if time.time() - timestamp > ttl_seconds:
        return None
    return payload.get("data")


def _save_cache(path: Path, data: object) -> None:
    try:
        path.parent.mkdir(parents=True, exist_ok=True)
        payload = {"timestamp": time.time(), "data": data}
        path.write_bytes(pickle.dumps(payload))
    except Exception as exc:
        logger.debug("Failed to write cache %s: %s", path, exc)


def _init_controller(config: Config, *, is_udm_pro: bool) -> UnifiController:
    from unifi_controller_api import UnifiController

    return UnifiController(
        controller_url=config.url,
        username=config.user,
        password=config.password,
        is_udm_pro=is_udm_pro,
        verify_ssl=config.verify_ssl,
    )


def fetch_devices(
    config: Config, *, site: str | None = None, detailed: bool = True
) -> Iterable[object]:
    """Fetch devices from UniFi Controller.

    Uses `unifi-controller-api` to authenticate and return device objects.
    """
    try:
        from unifi_controller_api import UnifiAuthenticationError
    except ImportError as exc:
        raise RuntimeError("Missing dependency: unifi-controller-api") from exc

    site_name = site or config.site
    ttl_seconds = _cache_ttl_seconds()
    cache_path = _cache_dir() / f"devices_{_cache_key(config.url, site_name, str(detailed))}.pkl"
    cached = _load_cache(cache_path, ttl_seconds)
    if cached is not None:
        logger.info("Using cached devices (%d)", len(cached))
        return cached

    try:
        controller = _init_controller(config, is_udm_pro=True)
    except UnifiAuthenticationError:
        logger.info("UDM Pro authentication failed, retrying legacy auth")
        controller = _init_controller(config, is_udm_pro=False)

    devices = controller.get_unifi_site_device(site_name=site_name, detailed=detailed, raw=False)
    _save_cache(cache_path, devices)
    logger.info("Fetched %d devices", len(devices))
    return devices


def fetch_clients(config: Config, *, site: str | None = None) -> Iterable[object]:
    """Fetch active clients from UniFi Controller."""
    try:
        from unifi_controller_api import UnifiAuthenticationError
    except ImportError as exc:
        raise RuntimeError("Missing dependency: unifi-controller-api") from exc

    site_name = site or config.site
    ttl_seconds = _cache_ttl_seconds()
    cache_path = _cache_dir() / f"clients_{_cache_key(config.url, site_name)}.pkl"
    cached = _load_cache(cache_path, ttl_seconds)
    if cached is not None:
        logger.info("Using cached clients (%d)", len(cached))
        return cached

    try:
        controller = _init_controller(config, is_udm_pro=True)
    except UnifiAuthenticationError:
        logger.info("UDM Pro authentication failed, retrying legacy auth")
        controller = _init_controller(config, is_udm_pro=False)

    clients = controller.get_unifi_site_client(site_name=site_name, raw=True)
    _save_cache(cache_path, clients)
    logger.info("Fetched %d clients", len(clients))
    return clients
